# This function builds a data package that maps internal HomoloGene
# ids of an organism to LocusLink ids, UniGene ids, percent identity
# of the alignment, and type of similarities of organisms of all
# pairwise best matches based on data from
# "ftp://ftp.ncbi.nih.gov/pub/HomoloGene/hmlg.ftp"
# pkgName - a character string for the name of the data package to be
# created.
# pkgPath - a character string for the path to which the data package
# to be created will be stored.
# version - a character string for the version number of the data
# package to be created.
# author - a list of two elements with the firest one being author - a
# character string for the name of the author of the data package and
# second one being maintainer - a character string for the email of
# the author.
#

homoPkgBuilder <- function(pkgName = "homology", pkgPath, version, author,
                 url = getSrcUrl("HG")){

    makeSrcInfo()
    createEmptyDPkg(pkgName, pkgPath)
    mapPS(url, pkgName, pkgPath)
    writeDescription(pkgName, pkgPath, version, author)
    # using pkgName() and the man page for the function
    writeFun (pkgPath, pkgName, organism = "")
    writeMan4Fun(pkgName, pkgPath, "hg")
    # Write .First.lib
    writeZZZ(pkgPath, pkgName)
    # Write the quality control data
    # Quality control
    getDPStats("", pkgName, pkgPath)
    writeMan4QC(pkgName, pkgPath)
    copySubstitute(file.path(.path.package("AnnBuilder"), "scripts",
                   "homoData.R"), file.path(pkgPath,
                   pkgName, "R", "homoData.R"), list(QQQQQQ = ""))
    copySubstitute(file.path(.path.package("AnnBuilder"), "scripts",
                   "homoData-class.Rd"), file.path(pkgPath, pkgName,
                   "man", "homoData-class.Rd"), list(QQQQQQ = ""))
    saveOrgNameNCode(pkgName, pkgPath, author)
    write("require(\"methods\") || stop(\"methods not available\")",
          file.path(pkgPath, pkgName, "install.R"))
    write("require(\"methods\") || stop(\"methods not available\")",
          file.path(pkgPath, pkgName, "R_PROFILE.R"))
}

procHomoData <- function(url = getSrcUrl("HG")){
    # scan() does not work for some reason
    homoData <- as.matrix(read.table(url, header = FALSE, sep = "|",
                           quote = "", as.is = TRUE, strip.white = TRUE,
                           comment.char = ""))
    # remove leading LL from LL ids
    homoData <- matrix(gsub("^LL\\.", "", homoData),
                       ncol = ncol(homoData), byrow = FALSE)
    # make "" NA
    homoData[homoData == ""] <- NA
    # remove leading white spaces
    homoData <- matrix(gsub(" *", "", homoData),
                       ncol = ncol(homoData), byrow = FALSE)
    return(homoData)
}

mapOrgs <- function(toMap, what = c("code", "name")){
    what <- match.arg(what)
    orgs <- getOrgNameNCode()
    if(what == "code"){
        return(orgs[toMap])
    }else{
        return(names(orgs[orgs == toMap]))
    }
}

getOrgNameNCode <- function(){
    return(c("3055" = "Chlamydomonas reinhardtii",
             "3702" = "Arabidopsis thaliana",
             "3847" = "Glycine max",
             "3880" = "Medicago truncatula",
             "4081" = "Lycopersicon esculentum",
             "4513" = "Hordeum vulgare",
             "4530" = "Oryza sativa",
             "4565" = "Triticum aestivum",
             "4577" = "Zea mays",
             "4896" = "Schizosaccharomyces pombe",
             "4932" = "Saccharomyces cerevisiae",
             "5833" = "Plasmodium falciparum",
             "6239" = "Caenorhabditis elegans",
             "7165" = "Anopheles gambiae",
             "7227" = "Drosophila melanogaster",
             "7719" = "Ciona intestinalis",
             "7955" = "Danio rerio",
             "8022" = "Oncorhynchus mykiss",
             "8090" = "Oryzias latipes",
             "8355" = "Xenopus laevis",
             "8364" = "Xenopus tropicalis",
             "9031" = "Gallus gallus",
             "9606" = "Homo sapiens",
             "9823" = "Sus scrofa",
             "9913" = "Bos taurus",
             "10090" = "Mus musculus",
             "10116" = "Rattus norvegicus",
             "44689" = "Dictyostelium discoideum"))
}

mapPS <- function(url, pkgName, pkgPath){
    homoMapping <- procHomoData(url)
    homoMapping <- rbind(homoMapping[, c(1, 5, 2, 3, 7, 8, 9, 10)],
                         homoMapping[, c(2, 8, 1, 3, 4, 5, 6, 10)])
    for(i in unique(homoMapping[, 1])){
        print(i)
        tempList <- list()
        tempHomoData <- HomoData2List(homoMapping[homoMapping[, 1] == i,
                              2:ncol(homoMapping)])
        saveList(tempHomoData, pkgName, pkgPath, i)
        tempList[["HOMODATA"]] <- i
        tempList[["ORGANISM"]] <- mapOrgs(i, "code")
        tempList[["BUILT"]] <- date()
        copySubstitute(file.path(.path.package("AnnBuilder"),
                                 "scripts", "homology#HOMODATA#.Rd"),
                           file.path(pkgPath, pkgName,
                           "man", paste(pkgName, i, ".Rd",
                                        sep = "")), tempList, "#")
    }
    ll2HGID <- cbind(homoMapping[, 5], paste(homoMapping[, 2],
                           mapOrgs(homoMapping[, 1], "code"), sep = "@"))
    saveMat(ll2HGID, pkgName, pkgPath, "LL2HGID", fun = twoStepSplit)
    writeManPage(pkgName, pkgPath, "LL2HGID", "",
                         src = get("LL2HGID", AnnInfo)$src)
}

HomoData2List <- function(data){
    data <- data[!is.na(data[,1]),]
    temp <- split.data.frame(data, factor(data[,1]))
    return(sapply(temp, getHomoDList))
}

getHomoDList <- function(data){
   return(apply(data, 1, getHomoData))
}

getHomoData <- function(entries){
    if(length(entries) < 7){
        stop("Incorrect argument length. Must be a vector of 7")
    }
    if(!is.na(entries[3]) && any(entries[3] == c("B", "b"))){
        return(new("homoData", homoOrg = mapOrgs(entries[2]),
                   homoLL = entries[4], homoType = entries[3],
                   homoPS = as.numeric(entries[7]),
                   homoACC = entries[6], homoHGID = entries[5],
                   homoURL = "NA"))
    }else{
        return(new("homoData", homoOrg = mapOrgs(entries[2]),
                   homoLL = entries[4], homoType = entries[3],
                   homoACC = entries[6], homoHGID = entries[5],
                   homoURL = as.character(entries[7])))
    }
}

saveOrgNameNCode <- function(pkgName, pkgPath, author){
    temp <- getOrgNameNCode()
    codeNName <- names(temp)
    names(codeNName) <- temp
    assign(paste(pkgName, "ORGCODE", sep = ""), codeNName)
    save(list = paste(pkgName, "ORGCODE", sep = ""),
	file =  file.path(pkgPath, pkgName, "data",
	paste(pkgName, "ORGCODE.rda", sep = "")))
    writeManPage(pkgName, pkgPath, "ORGCODE", "",
                     src = get("ORGCODE", AnnInfo)$src)
}

getLL2IntID <- function(homoData, organism = ""){
    switch(tolower(organism),
           human = org <- "9606",
           mouse = org <- "10090",
           rat = org <- "10116",
           org <- NULL)
    ll2ID <- homoData[, c(4, 5, 1)]
    ll2ID <- rbind(ll2ID, homoData[,c(7, 8, 2)])
    if(!is.null(org)){
        ll2ID <- ll2ID[ll2ID[,3] == org, ]
    }
    ll2ID <- ll2ID[, 1:2]
    # Drop entries without LL
    ll2ID <- ll2ID[ll2ID[, 1] != "NA", ]
    # Get duplicated LL
    if(nrow(ll2ID) > length(unique(ll2ID[,1]))){
        ll2ID <- mergeRowByKey(ll2ID)
    }
    colnames(ll2ID) <- c("LOCUSID", "HGID")

    return(ll2ID)
}

